package de.isse.kiv

import java.io.File
import org.eclipse.core.runtime.preferences.AbstractPreferenceInitializer
import org.eclipse.jface.preference.IPreferenceStore
import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.ui.Console
import java.io.InputStream
import java.io.FileOutputStream
import org.eclipse.swt.graphics.Color
import org.eclipse.swt.graphics.RGB
import org.eclipse.jface.preference.PreferenceConverter
import de.isse.kiv.Settings._

object Settings {
  val CONSOLE_OUT = "de.isse.kiv.preferences.out"
  val CONSOLE_DEBUG = "de.isse.kiv.preferences.debug"
  val CONSOLE_ERROR = "de.isse.kiv.preferences.error"

  val KIVHOME = "de.isse.kiv.preferences.kivhome"
  val FORK = "de.isse.kiv.preferences.fork"
  val EXPERT = "de.isse.kiv.preferences.expert"
  val INTERNAL_EDITOR = "de.isse.kiv.preferences.internaleditor"
  val STACKSIZE = "de.isse.kiv.preferences.stacksize"

  val PROJECT_LOCATIONS = "de.isse.kiv.preferences.projectlocations"
  val PROJECT_RECENT = "de.isse.kiv.preferences.projectrecent"

  val HIGHLIGHT_VARIABLES = "de.isse.kiv.preferences.highlightvariables"
  val SYNTAX_BOLD = ".bold"
  val SYNTAX_BOLD_STRING = "Bold"
  val SYNTAX_ITALIC = ".italic"
  val SYNTAX_ITALIC_STRING = "Italic"
  val SYNTAX_STRIKETHROUGH = ".strikethrough"
  val SYNTAX_STRIKETHROUGH_STRING = "Strikethrough"
  val SYNTAX_UNDERLINED = ".underlined"
  val SYNTAX_UNDERLINED_STRING = "Underlined"
  val SYNTAX_COLOR = ".color"
  val SYNTAX_COLOR_STRING = "Color:"

  val KEYWORD = "keyword"
  val PROGRAM_CONSTRUCT = "program construct"
  val PROCEDURE = "procedure"
  val SORT = "sort"
  val VARIABLE = "variable"
  val OP = "op"
  val OPERATOR = "operator"
  val COMMENT = "comment"
  val STRING = "string"
  val NUMBER = "number"
  val THEOREM = "theorem"
  val SPECIFICATION = "specification"

  val TOKEN_NAME_MAP = Map(
    KEYWORD -> "de.isse.kiv.preferences.syntaxcoloring.keyword",
    PROGRAM_CONSTRUCT -> "de.isse.kiv.preferences.syntaxcoloring.program_construct",
    PROCEDURE -> "de.isse.kiv.preferences.syntaxcoloring.procedure",
    SORT -> "de.isse.kiv.preferences.syntaxcoloring.sort",
    VARIABLE -> "de.isse.kiv.preferences.syntaxcoloring.variable",
    OP -> "de.isse.kiv.preferences.syntaxcoloring.op",
    OPERATOR -> "de.isse.kiv.preferences.syntaxcoloring.bold_op",
    COMMENT -> "de.isse.kiv.preferences.syntaxcoloring.comment",
    STRING -> "de.isse.kiv.preferences.syntaxcoloring.string",
    NUMBER -> "de.isse.kiv.preferences.syntaxcoloring.number",
    THEOREM -> "de.isse.kiv.preferences.syntaxcoloring.theorem",
    SPECIFICATION -> "de.isse.kiv.preferences.syntaxcoloring.specification")

  val KIVHOME_ENVIRON = "$KIVHOME"
  val KIVHOME_DEFAULT = "$PLUGIN_LIBRARY"
  val STACKSIZE_DEFAULT = "16m"

  val EDITOR_BRACKET_MATCH_PAIRS = Array('{', '}', '(', ')', '[', ']', '⟨', '⟩', '⦉', '⦊')
  val EDITOR_MATCHING_BRACKETS = "de.isse.kiv.settings.editor.matchingBrackets"
  val EDITOR_MATCHING_BRACKETS_COLOR = "de.isse.kiv.settings.editor.matchingBracketsColor"
  val EDITOR_MATCHING_BRACKETS_COLOR_VALUE = "128,0,128"

  private def getStringPreference(pref: String): String = {
    KIVPlugin.getDefault.getPreferenceStore.getString(pref)
  }

  private def setStringPreference(pref: String, value: String) {
    KIVPlugin.getDefault.getPreferenceStore.setValue(pref, value)
  }

  def kivHome: String = {
    val value = getStringPreference(KIVHOME)
    if (value == KIVHOME_DEFAULT) {
      try {
        return ResourceLookup.pluginPath(KIVSystem.LIBRARY_PLUGIN_ID).toOSString()
      } catch {
        case e: Exception =>
      }
    }
    if (value == KIVHOME_ENVIRON || value == KIVHOME_DEFAULT) {
      val kivHome = System.getenv("KIVHOME")
      if (kivHome != null) return kivHome else return ""
    }
    value
  }

  def isForkRequired: Boolean = {
    KIVPlugin.getDefault.getPreferenceStore.getBoolean(FORK)
  }

  def isExpertMode: Boolean = {
    KIVPlugin.getDefault.getPreferenceStore.getBoolean(EXPERT)
  }

  def useInternalEditor: Boolean = {
    KIVPlugin.getDefault.getPreferenceStore.getBoolean(INTERNAL_EDITOR)
  }

  def doHighlightVariables: Boolean = {
    KIVPlugin.getDefault.getPreferenceStore.getBoolean(HIGHLIGHT_VARIABLES)
  }

  def stacksize: String = getStringPreference(STACKSIZE)

  def hasProjectLocation(path: String) = {
    projectLocations contains path
  }

  def projectLocations: Array[String] = {
    val paths = getStringPreference(PROJECT_LOCATIONS)
    if (paths.isEmpty)
      Array()
    else
      paths.split(File.pathSeparator)
  }

  def projectLocations_=(locations: Array[String]) = {
    val paths = locations.mkString(File.pathSeparator)
    setStringPreference(PROJECT_LOCATIONS, paths)
  }

  def addProjectLocation(path0: String) {
    val path = path0.trim()
    if (!path.isEmpty) {
      val locations = projectLocations
      if (!(locations contains path))
        projectLocations = Array(path) ++ locations
    }
  }

  def removeProjectLocation(path0: String) {
    val path = path0.trim()
    if (!path.isEmpty) {
      val locations = projectLocations
      projectLocations = locations.filter(_ != path)
    }
  }

  def touchProjectLocation(path: String) {
    setStringPreference(PROJECT_RECENT, path)
  }

  def recentProjectLocation = {
    getStringPreference(PROJECT_RECENT)
  }
}

class Settings extends AbstractPreferenceInitializer {
  def initializeDefaultPreferences() {

    val plugin = KIVPlugin.getDefault
    val preferences = plugin.getPreferenceStore

    preferences.setDefault(Settings.CONSOLE_OUT, true)
    preferences.setDefault(Settings.CONSOLE_DEBUG, false)
    preferences.setDefault(Settings.CONSOLE_ERROR, true)
    val kivHome = System.getenv("KIVHOME")
    preferences.setDefault(Settings.KIVHOME, if (kivHome == null) Settings.KIVHOME_DEFAULT else Settings.KIVHOME_ENVIRON)
    preferences.setDefault(Settings.STACKSIZE, Settings.STACKSIZE_DEFAULT)
    preferences.setDefault(Settings.FORK, true)
    preferences.setDefault(Settings.EXPERT, false)
    preferences.setDefault(Settings.INTERNAL_EDITOR, true)
    // preferences.setDefault(Settings.EDITOR_MATCHING_BRACKETS, true)
    // preferences.setDefault(Settings.EDITOR_MATCHING_BRACKETS_COLOR, Settings.EDITOR_MATCHING_BRACKETS_COLOR_VALUE)

    //Colors
    preferences.setDefault(TOKEN_NAME_MAP.get(KEYWORD).get + SYNTAX_COLOR, "127, 0, 85")
    preferences.setDefault(TOKEN_NAME_MAP.get(PROGRAM_CONSTRUCT).get + SYNTAX_COLOR, "0, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(PROCEDURE).get + SYNTAX_COLOR, "0, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(SORT).get + SYNTAX_COLOR, "192, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(VARIABLE).get + SYNTAX_COLOR, "0, 0, 192")
    preferences.setDefault(TOKEN_NAME_MAP.get(OP).get + SYNTAX_COLOR, "0, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(OPERATOR).get + SYNTAX_COLOR, "0, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(COMMENT).get + SYNTAX_COLOR, "0, 128, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(STRING).get + SYNTAX_COLOR, "255, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(NUMBER).get + SYNTAX_COLOR, "0, 0, 0")
    preferences.setDefault(TOKEN_NAME_MAP.get(THEOREM).get + SYNTAX_COLOR, "0, 0, 100")
    preferences.setDefault(TOKEN_NAME_MAP.get(SPECIFICATION).get + SYNTAX_COLOR, "0, 0, 192")

    //Bold Tokens
    preferences.setDefault(TOKEN_NAME_MAP.get(KEYWORD).get + SYNTAX_BOLD, true)
    preferences.setDefault(TOKEN_NAME_MAP.get(PROGRAM_CONSTRUCT).get + SYNTAX_BOLD, true)
    preferences.setDefault(TOKEN_NAME_MAP.get(PROCEDURE).get + SYNTAX_BOLD, true)
    preferences.setDefault(TOKEN_NAME_MAP.get(OPERATOR).get + SYNTAX_BOLD, true)
    preferences.setDefault(TOKEN_NAME_MAP.get(THEOREM).get + SYNTAX_BOLD, true)
    preferences.setDefault(TOKEN_NAME_MAP.get(SPECIFICATION).get + SYNTAX_BOLD, true)

    //Underlined Tokens
    preferences.setDefault(TOKEN_NAME_MAP.get(THEOREM).get + SYNTAX_UNDERLINED, true)

    preferences.setDefault(Settings.HIGHLIGHT_VARIABLES, true)
  }

  def ColorToRGB(color: Color): RGB = {
    new RGB(color.getRed(), color.getGreen(), color.getBlue())
  }
}
