package de.isse.kiv.project

import de.isse.kiv.source.parser.KIVParser
import de.isse.kiv.ui.Console
import de.isse.kiv.resources.ResourceProperties._
import org.eclipse.core.resources.IncrementalProjectBuilder
import org.eclipse.core.runtime.IProgressMonitor
import org.eclipse.core.resources.IFile
import org.eclipse.core.resources.IProject
import org.eclipse.core.runtime.CoreException
import org.eclipse.core.runtime.Path
import kiv.util.MultiGraph
import kiv.util.DefaultEdge
import de.isse.kiv.KIVEditorServer
import de.isse.kiv.resources.FileList
import de.isse.kiv.resources.Marker
import de.isse.kiv.resources.ProjectModel
import de.isse.kiv.resources.ResourceLookup
import de.isse.kiv.resources.ResourceProperties

object Builder {
  val BUILDER_ID = "de.isse.kiv.resources.builder"
}

class Builder extends IncrementalProjectBuilder {
  private def rebuildFile(file: IFile, monitor: IProgressMonitor) {
    try {
      Console.debug.println("build file", file.getName)
      monitor.subTask("building file " + file.getName)

      val stream = file.getContents
      val text = scala.io.Source.fromInputStream(stream, "UTF-8").mkString("")
      KIVParser.parse(text, file)
    } catch {
      case e: CoreException =>
    } finally {
      monitor.worked(1)
    }
  }

  override def clean(monitor: IProgressMonitor) {
    Marker.deleteMarkers(getProject, Marker.ERROR_MARKER_ID)
    Marker.deleteMarkers(getProject, Marker.WARNING_MARKER_ID)
    Marker.deleteMarkers(getProject, Marker.TEXT_MARKER_ID)
  }

  private def buildAll(model: ProjectModel, specs: Map[String, IFile], monitor: IProgressMonitor) {
    var todo: List[String] = specs.keys.toList
    var done: Set[String] = Set()

    while (!todo.isEmpty) {
      val name = todo.head
      if (!(done contains name) && (specs contains name)) {
        rebuildFile(specs(name), monitor)
        todo = model.subSpecs.getOrElse(name, Nil) ++ todo.tail
        done += name
      }
    }
  }

  /* build files in the changeset and their dependencies (reachable)
   * in topological order.
   */
  protected def build(kind: Int, args: java.util.Map[String, String], monitor: IProgressMonitor): Array[IProject] = try {
    val full = {
      kind == IncrementalProjectBuilder.FULL_BUILD || kind == IncrementalProjectBuilder.CLEAN_BUILD
    }

    val project = getProject
    val specsPath = new Path(ResourceProperties.SPECS)
    val model = project.projectModel

    val files =
      if (full | model.needsFullBuild)
        FileList.collect(project.getFolder(specsPath))
      else {

        FileList.collectDelta(getDelta(project).findMember(specsPath))
      }

    val kivfiles = files.filter { _.getName.endsWith(".utf8") }
    val changed = kivfiles.map(_.specName)

    monitor.beginTask("build project " + project.getName, kivfiles.size)
    Console.debug.println("build project", project.getName)

    val subspecs = model.subSpecs
    val graph = new MultiGraph[String, DefaultEdge[String]]()

    // first use of graph: compute reachable dependencies

    val nodes = subspecs.keySet ++ changed // Note: DO include changed, otherwise graph will just fail in reachableNodes

    val edges = for ((sup, subs) <- subspecs; sub <- subs)
      yield new DefaultEdge(sub, sup) // edges point upwards the spec hierarchy

    nodes foreach graph.add
    edges foreach graph.add

    val reachable = graph reachableNodes changed

    // second use of graph: compute topsort on reachable nodes

    val unreachable = nodes -- reachable
    graph removeNodes unreachable

    val sorted = try {
      graph.topological_sort()
    } catch {
      case _: Exception => reachable // in case of cycle
    }

    for (spec <- sorted) {
      if (!monitor.isCanceled) {
        rebuildFile(ResourceLookup.specFile(spec, project), monitor)
        rebuildFile(ResourceLookup.seqFile(spec, project), monitor)
      }
    }

    monitor.done()
    null
  } catch {
    case e: Exception =>
      Console.error.println("build project", getProject.getName, "failed")
      Console.error.printShort(e)
      null
  }
}
