package de.isse.kiv.project

import org.eclipse.core.resources.IFile
import kiv.prog.AnyProc
import de.isse.kiv.ui.Console
import de.isse.kiv.resources.ResourceLookup._
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.ui.editors.FileEditor._
import kiv.expr.Op
import kiv.printer.prettyprint._
import de.isse.kiv.source.parser._
import kiv.parser.PreSort
import kiv.parser.PreTyCo
import kiv.expr.TyCo

object Definition {
  /**
   * Get an action that opens the specification file that contains the declaration of the procedure.
   * Additionally, jumps to/selects the name of the declaration.
   */
  def gotoDeclarationAction(file: IFile, proc: AnyProc): Option[() => Unit] = {
    file.getProject.projectModel.getDeclaration(file.specName, proc) match {
      case Some((decl, specname)) =>
        val specFile = specFileOrLibrary(specname, file.getProject)
        if (specFile.fileExists)
          Some({ () =>
            val filemodel = specFile.fileModel
            val range = if (filemodel != null) filemodel.findDefinitionRange(decl.declprocdecl.proc) else None
            openFile(specFile, range)
          })
        else {
          Console.error.println(s"File ${specFile} does not exist")
          None
        }
      case _ =>
        Console.error.println(s"Could not find declaration for ${proc.procsym.name}")
        None
    }
  }

  /**
   * Get an action that opens the specification file that contains the definition of the operation.
   * Additionally, jumps to/selects the first occurrence of the operation
   */
  def gotoDefinitionAction(file: IFile, op: Op): Option[() => Unit] = {
    val projectModel = file.getProject.projectModel
    projectModel.getDefinition(file.specName, op) match {
      case Some((origop, filename)) =>
        val fileToOpen = specFileOrLibrary(filename, file.getProject)
        if (!fileToOpen.fileExists) {
          Console.error.println(s"File ${fileToOpen} does not exist")
          return None
        }

        val range = fileToOpen match {
          case Left(fileToOpen) =>
            val fileModel = fileToOpen.fileModel
            if (fileModel != null)
              fileModel.findDefinitionRange(origop, fileToOpen)
            else
              None
          case Right(_) =>
            None
        }
        Some({ () =>
          openFile(fileToOpen, range)
        })
      case None =>
        None
    }
  }

  /**
   * Get an action that opens the specification file that contains the definition of the sort.
   * Additionally, jumps to/selects the first occurrence of the operation
   * */
  def gotoDefinitionAction(file: IFile, tyco: PreTyCo): Option[() => Unit] = {
    file.getProject.projectModel.getDefinition(file.specName, TyCo(tyco.pretycosymloc.sym, tyco.pretycoarity)) match {
      case Some((origtyco, specname)) =>
        val specFile = specFileOrLibrary(specname, file.getProject)
        if (specFile.fileExists)
          Some({ () =>
            val filemodel = specFile.fileModel
            val range = if (filemodel != null) filemodel.findDefinitionRange(origtyco) else None
            openFile(specFile, range)
          })
        else {
          Console.error.println(s"File ${specFile} does not exist")
          None
        }
      case _ =>
        Console.error.println(s"Could not find declaration for ${tyco.pretycosymloc.sym.name}")
        None
    }
  }

  def gotoDefinitionAction(file: IFile, token: KIVToken): Option[() => Unit] = token match {
    case ProcedureToken(preproc) =>
      Definition.gotoDeclarationAction(file, preproc.proc)
    case OperationToken(op) =>
      Definition.gotoDefinitionAction(file, op.op)
    case SortToken(tyco) =>
      Definition.gotoDefinitionAction(file, tyco)
    case _ =>
      None
  }
}
