package de.isse.kiv.source

import scala.Array.canBuildFrom
import scala.annotation.tailrec
import scala.collection.mutable.ArrayBuffer
import org.eclipse.jface.text.IDocument
import org.eclipse.jface.text.ITextViewer
import org.eclipse.jface.text.Region
import org.eclipse.jface.text.contentassist.CompletionProposal
import org.eclipse.jface.text.contentassist.ICompletionProposal
import org.eclipse.jface.text.contentassist.IContentAssistProcessor
import org.eclipse.jface.text.contentassist.IContextInformation
import org.eclipse.jface.text.contentassist.IContextInformationValidator
import org.eclipse.jface.text.templates.DocumentTemplateContext
import org.eclipse.jface.text.templates.Template
import org.eclipse.jface.text.templates.TemplateContext
import org.eclipse.jface.text.templates.TemplateContextType
import org.eclipse.jface.text.templates.TemplateProposal
import org.eclipse.ui.texteditor.ITextEditor
import kiv.converter.KivFont
import kiv.expr._
import kiv.printer.prettyprint.xpp
import kiv.prog._
import de.isse.kiv.ui.Console
import kiv.expr.Xov
import scala.collection.mutable.ListBuffer
import scala.reflect.internal.util.Collections
import de.isse.kiv.KIVPlugin
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.refactoring.KIVRefactoring

class SymbolProcessor(val editor: ITextEditor) extends IContentAssistProcessor with WithSpec {
  private val NO_CONTEXTS: Array[IContextInformation] = Array()

  private val PROPOSAL_ACTIVATION_CHARS = Array('\\', '.')

  private val NO_PROPOSALS: Array[(String, String)] = Array()
  private val NO_COMPLETIONS: Array[ICompletionProposal] = Array()

  val completions = List(("\\ ", "\\")) ++ {
    for (
      kc <- KivFont.getKivChars if kc != null && kc.names != null;
      key <- kc.names
    ) yield (key, kc.unicode.toString)
  }

  private def findOpsAndProcsProposals(document: IDocument, _offset: Int, _length: Int) = {
    val _spec = spec

    //Sonderfall Anfang des Dokumentes (leerzeichen und dann zeichenfolge führt sonst zu problemen)
    val offset = if (_offset != 0 || (document.getChar(0).isWhitespace)) _offset else _offset - 1
    val length = if (_offset != 0 || (document.getChar(0).isWhitespace)) _length else _length + 1

    val filter = document.get(offset + 1, length - 1)

    //Do not insert something when only whitespace and ctrl+space
    if (filter.length == 0 || _spec == None) {
      Nil
    } else {
      val sig = _spec.get.specsignature

      //extract Ops proposals with prioint == 0
      val ops = sig.oplist.collect {
        case Op(Symbol(compl), typ @ Funtype(argtypes, _), 0, _) if compl.startsWith(filter) =>
          (compl, typ, argtypes)
      }

      //build Ops CompletionProposals 
      val ops_proposals = ops.map {
        case (compl, typ, argtypes) =>
          val vars = KIVRefactoring.newVars(argtypes, Nil, spec.get.specvars)
          val names = vars.map("${" + _.xovstring + "}")
          val pattern = compl + "(" + names.mkString(", ") + ")"        
       
          val template = new Template(compl, xpp(typ), "stdType", pattern, true)
          val contextType = new TemplateContextType("stdType")
          val ctx = new DocumentTemplateContext(contextType, document, offset + 1, length - 1)
          val region = new Region(offset + 1, length - 1)

          new TemplateProposal(template, ctx, region, null) {
            // default implementation puts an unwanted "-" in between
            override def getDisplayString = compl + " " + xpp(typ)
          }
      }

      //Procs Proposals
      val procs = projectModel.getProcCompletions(file, filter).map {
        case proc @ NewProc(Symbol(compl), mode, _, _, _) =>
          val declfpl = projectModel.getReducedDeclaration(file, proc)
          val params = declfpl match {
            case Some((_, Fpl(values, vars, outs))) =>
              val valp = values.map("${" + xpp(_) + "}")
              val varp = vars.map("${" + xpp(_) + "}")
              val outp = outs.map("${" + xpp(_) + "}")
              buildParamString(valp, varp, outp)

            case None =>

              val valueparams = mode.mvalueparams
              val varparams = mode.mvarparams
              val outparams = mode.moutparams

              //Get unique variables for each datatype
              val dataTypesNames = (valueparams ++ varparams ++ outparams).distinct.map(t => t -> getUniqueVar(t)).toMap
              val dataTypesCounter = scala.collection.mutable.Map(((valueparams ++ varparams ++ outparams).distinct.map(t => (t -> 0)): _*))

              val valp = valueparams.map(
                {
                  typ =>
                    val counter = dataTypesCounter.get(typ).get
                    dataTypesCounter.update(typ, counter + 1)

                    "${" + dataTypesNames.get(typ).get + counter + "}"

                })

              val varp = varparams.map(
                {
                  typ =>
                    val counter = dataTypesCounter.get(typ).get
                    dataTypesCounter.update(typ, counter + 1)

                    "${" + dataTypesNames.get(typ).get + counter + "}"
                })

              val outp = outparams.map(
                {
                  typ =>
                    val counter = dataTypesCounter.get(typ).get
                    dataTypesCounter.update(typ, counter + 1)

                    "${" + dataTypesNames.get(typ).get + counter + "}"
                })

              buildParamString(valp, varp, outp)

          }
          val shortcompl = projectModel.getASMName(file) match {
            case Some(name) =>
              if (compl.startsWith(name.name)) compl.substring(name.name.length + 1)
              else compl
            case None =>
              compl
          }
          (shortcompl, xpp(mode), params)
      }

      //Build Procs TemplateProposals
      val procs_templates = procs.map {
        case (replace, typ, pattern) =>
          val _typ = typ.replace("\n", "")
          val template = new Template(replace, _typ, "", (replace + pattern), true)
          val contextType = new TemplateContextType("stdType")
          val ctx = new DocumentTemplateContext(contextType, document, offset + 1, length - 1)
          val region = new Region(offset + 1, length - 1)

          new TemplateProposal(template, ctx, region, null) {
            // default implementation puts an unwanted "-" in between
            override def getDisplayString = replace + " " + _typ
          }
      }

      //Build Procs Completion Proposals (no variables)
      val procs_proposals = procs.map {
        case (replace, typ, _) =>
          new CompletionProposal(replace, offset + 1, length - 1, replace.length, null, replace, null, null)
      }

      ops_proposals ++ procs_proposals ++ procs_templates
    }
  }

  /*Returns a unique flexible variable to a given type
   * returns _ if no unique flexible variable is available
   */
  def getUniqueVar(typ: kiv.expr.Type): String = {
    val specvars = spec.get.specvars.filter { x => x.typ == typ && x.flexiblep }

    if (specvars.nonEmpty) {
      val varNames = specvars.map(_.xovstring)

      //Check if flexible vars are unique
      val prefix = varNames.head.split(Array('0', '1', '2', '3', '4', '5', '6', '7', '8', '9')).head
      val notUnique = varNames.find { x => !(x.startsWith(prefix) && (x.length == prefix.length || x.splitAt(prefix.length())._2.charAt(0).isDigit)) }

      if (notUnique.isEmpty)
        return prefix
    }

    //Placeholder
    "_" + xpp(typ)
  }

  def buildParamString(valp: List[String], varp: List[String], outp: List[String]): String = {
    val s0 = valp.mkString(", ")
    val s1 = if (varp.isEmpty && outp.isEmpty) "" else "; " + varp.mkString(", ")
    val s2 = if (outp.isEmpty) "" else "; " + outp.mkString(", ")
    "(" + s0 + s1 + s2 + ")"

  }
  @tailrec
  private def searchBackwards(document: IDocument, offset: Int, delimChars: Array[Char]): Option[(Int, Char)] = {
    if (offset < 0) Some((0, ' '))
    else if ((delimChars contains document.getChar(offset)) || document.getChar(offset).isWhitespace) Some((offset, document.getChar(offset)))
    else searchBackwards(document, offset - 1, delimChars)
  }

  private def findSymbolProposals(document: IDocument, bspos: Int, length: Int) = {
    // remove backspace at bspos
    val range = document.get(bspos + 1, length - 1)

    val symbols = completions.filter {
      case (key, replace) => key startsWith range
    }

    symbols.map {
      case (key, replace) =>
        new CompletionProposal(replace, bspos, length, replace.length, null, key + " " + replace, null, null)
    }
  }

  private def findSignatureProposals(document: IDocument, bspos: Int, length: Int) = {
    //do not remove dot
    val range = document.get(bspos, length)

    val selectors = spec.toList.flatMap { _.specfcts.collect {
      case op @ Op(Symbol(name), _, _, _) if name.startsWith(range) && !name.endsWith(":=") =>
        op
    }}

    val result = selectors.collect {
      case op @ Op(_, Funtype(_, res), _, _) if res.sortp =>
        val name = op.opstring
        val typ = res.toSort.sortstring
        new CompletionProposal(name, bspos, length, name.length, null, name + ": " + typ, null, null)
    }

    result
  }

  private def findProposals(document: IDocument, bspos: Int, length: Int) = {
    if (document.getChar(bspos) == '\\')
      findSymbolProposals(document, bspos, length)
    else if (document.getChar(bspos) == '.')
      findSignatureProposals(document: IDocument, bspos: Int, length: Int)
    else
      Nil
  }

  def computeCompletionProposals(viewer: ITextViewer, _offset: Int): Array[ICompletionProposal] = {
    //TODO Template lassen sich nicht mit Leerzeichen einfügen
    //Completions schon. Ist das ein Bug? Diese methode wird bei Templates nicht mehr aufgerufen
    //da die Auswahl aber mit Pfeiltasten erfolgt, drückt man danach eigentlich Enter? Also vllt kein problem

    try {
      val document = viewer.getDocument

      var isSpace = false;
      var offset = _offset;
      if (_offset != 0) {
        isSpace = (document.getChar(_offset - 1) == ' ')
        offset = if (isSpace) _offset - 1 else _offset
      }

      val (start, character) = searchBackwards(document, offset - 1, PROPOSAL_ACTIVATION_CHARS).getOrElse(return NO_COMPLETIONS)

      val length = offset - start
      var proposals: List[ICompletionProposal] = findProposals(document, start, length)

      if (character.isWhitespace) {
        //_offset - start - (_offset-offset) strips the last whitespace
        var shortedLength = _offset - start - (_offset - offset)
        proposals = proposals ++ findOpsAndProcsProposals(document, start, shortedLength)

      }
      val isDot = (document.getChar(start) == '.')

      val doComplete = !isDot || length > 1
      if (isSpace) {
        if (doComplete) {
          proposals.headOption.map(_.apply(document))
        }

        return NO_COMPLETIONS
      } else {
        return proposals.toArray.sortBy(_.getDisplayString)
      }
    } catch {
      case e: Exception =>
        Console.error.println("completion failed")
        Console.error.printShort(e)
        NO_COMPLETIONS
    }
  }

  def computeContextInformation(viewer: ITextViewer, offset: Int): Array[IContextInformation] = NO_CONTEXTS

  def getCompletionProposalAutoActivationCharacters(): Array[Char] = PROPOSAL_ACTIVATION_CHARS

  override def getContextInformationAutoActivationCharacters(): Array[Char] = null

  override def getErrorMessage(): String = null

  override def getContextInformationValidator(): IContextInformationValidator = null
}
