package de.isse.kiv.source.parser

import org.eclipse.core.resources.IFile
import org.eclipse.core.resources.IResource
import de.isse.kiv.resources.Marker
import de.isse.kiv.resources.ResourceProperties._
import de.isse.kiv.ui.Console
import kiv.basic._
import kiv.parser._
import kiv.spec._
import de.isse.kiv.source.analyzers._

object KIVParser {
  def parse(text: String, file: IFile): Any = {
    val parsedText = try {
      Marker.deleteMarkers(file, Marker.ERROR_MARKER_ID)
      Marker.deleteMarkers(file, Marker.WARNING_MARKER_ID)
      parseText(text, file)
    } catch {
      case error: KIVError =>
        parseError(error, file)
      case e: Exception =>
        Console.debug.println("unknown exception during parsing: ")
        Console.debug.print(e)
    }
    return parsedText
  }

  private def parseText(text: String, file: IFile) = {
    val fileName = file.getName
    val project = file.getProject
    val projectModel = project.projectModel
    val fileModel = file.fileModel
    val specName = file.specName
    val signatureBefore = projectModel.signatureBefore(file)

    Console.debug.println("parsing", specName + "/" + fileName)

    val parsed: Any = {
      if (fileName == "specification.utf8") {
        val (prespec, tospec) = Parse.parse_prespec(text, projectModel.specList)
        val spec = tospec()
        Console.debug.println("updating " + specName + " model")
        projectModel.updateSpec(specName, spec)
        val warningList = Analyzer.checkSpecification(spec)
        warningList.foreach { warning =>
          Marker.createWarningMarker(file, warning.text, warning.text,
            warning.location.line, warning.location.position, warning.location.position + warning.location.length)
        }
        fileModel.update(spec)
        prespec 
      } else if (fileName == "sequents.utf8") {
        val theorems = Parse.parse_theoremlist(text, signatureBefore)
        val warningList = Analyzer.checkSequents(theorems)
        warningList.foreach { warning =>
          Marker.createWarningMarker(file, warning.text, warning.text,
            warning.location.line, warning.location.position, warning.location.position + warning.location.length)
        }
        theorems
      } else if (fileName == "patterns.utf8") {
        Parse.parse_patterns(text, signatureBefore)
      } else {
        val errormsg = "unknown file type: " + fileName
        Marker.createErrorMarker(file, errormsg, errormsg, 0, 0, 0)
      }
    }
    // TODO: checks
    val procMap = projectModel.getProcMap(file)
    
    val specAnalyzer: Option[SpecMapsAnalyzer] = parsed match {
      case prespec: PreSpecification => Some(new PreSpecificationAnalyzer(prespec, procMap))
      case theoremlist: List[Theorem @unchecked] if theoremlist.forall(_.isInstanceOf[Theorem]) => Some(new SequentsAnalyzer(theoremlist, procMap))
      case _ => None
    }
    specAnalyzer.map { analyzer => file.fileModel.updateSpecAnalyzer(analyzer) }

    fileModel.update(parsed)
    parsed
  }

  private def parseError(error: KIVError, file: IResource) {
    val location = error match {
      case error: KIVError with SourceLocation if error.location != None =>
        error.location.get
      case _ =>
        Location(1, 0, 0, 0)
    }
    error match {
      case _: Typeerror | _: Parsererror | _: Signatureerror =>
      case e =>
        Console.debug.println("unexpected error during parsing with stacktrace:")
        Console.debug.print(e)
    }
    val errormsg = error.errorstringlist.mkString("\n").trim + " (" + error.typeOfError + ")"
    Marker.createErrorMarker(file, errormsg, errormsg, location.line, location.position, location.position + location.length)
  }
}
