package de.isse.kiv.source.parser

import kiv.parser._
import kiv.prog._
import kiv.expr._
import kiv.printer.prettyprint.xpp
import kiv.parser.Terminals._
import org.eclipse.jface.text.rules.IToken
import de.isse.kiv.source.TokenMap

sealed abstract class KIVToken {
  val location: Location
  val eclipseToken: IToken
  def stringPresentation: String
  override def toString: String = {
    stringPresentation + " " + location.position + " " + location.length
  }

  /** True if we should mark the occurrences of this token type */
  val markOccurrences: Boolean = false
  def markOccurrence(marker: KIVToken): Boolean = false

  /** True if the rename refactoring can be applied to it */
  val canRename: Boolean = false
}

case class RawToken(val terminal: Terminal, val tokenString: String, override val location: Location) extends KIVToken { // common case
  def this(terminal: Terminal, tokenString: String, range: Range) = {
    this(terminal, tokenString, new Location(terminal.getLine, terminal.getColumn, range.start, range.length))
  }
  override val eclipseToken: IToken = {
    terminal.getId() match {
      /*case T_XOV | T_PROC | T_XOV_SYM | T_SORT_OR_XOV | T_FCT | /*T_INFIX* |*/ T_PRAEFIXFCT | T_PRAEFIXPRD | T_METHODFCT | T_POSTFIXFCT
      | T_INFIXFCTL1 | T_INFIXFCTL2 | T_INFIXFCTL3 | T_INFIXFCTL4 | T_INFIXFCTL5 | T_INFIXFCTL6 | T_INFIXFCTL7 | T_INFIXFCTL8
      | T_INFIXFCTL9 | T_INFIXFCTL10 | T_INFIXFCTL11 | T_INFIXFCTL12 | T_INFIXFCTL13 | T_INFIXFCTL14 | T_INFIXFCTL15 | T_INFIXFCTR1
      | T_INFIXFCTR2 | T_INFIXFCTR3 | T_INFIXFCTR4 | T_INFIXFCTR5 | T_INFIXFCTR6 | T_INFIXFCTR7 | T_INFIXFCTR8 | T_INFIXFCTR9
      | T_INFIXFCTR10 | T_INFIXFCTR11 | T_INFIXFCTR12 | T_INFIXFCTR13 | T_INFIXFCTR14 | T_INFIXFCTR15 =>
        TokenMap.ERROR*/
      case _ =>
        TokenMap.get(terminal.getId())
    }
  }
  override def stringPresentation: String = {
    tokenString
  }
}

case class SortToken(val pretyco: PreTyCo) extends KIVToken {
  override val location = pretyco.pretycosymloc.loc
  override val eclipseToken: IToken = TokenMap.SORT
  override def stringPresentation: String = {
    pretyco.pretycosymloc.sym.name
  }
  override val markOccurrences: Boolean = true
  override def markOccurrence(marker: KIVToken): Boolean = marker match {
    case _: SortToken | _: SortDefinitionToken =>
      stringPresentation == marker.stringPresentation
    case _ =>
      false
  }
}

case class SortDefinitionToken(val tycodef: TyCodef) extends KIVToken {
  override val location = tycodef.symloc.loc
  override val eclipseToken: IToken = TokenMap.SORT
  override def stringPresentation: String = {
    tycodef.symloc.sym.name
  }
  override val markOccurrences: Boolean = true
  override def markOccurrence(marker: KIVToken): Boolean = marker match {
    case _: SortToken | _: SortDefinitionToken =>
      stringPresentation == marker.stringPresentation
    case _ =>
      false
  }
}

case class OperationToken(val op: PreOp) extends KIVToken {
  override val location: Location = op.location.get
  override val eclipseToken: IToken = TokenMap.OP
  override def stringPresentation: String = {
    op.opsym.name
  }
  override val markOccurrences: Boolean = true
  override def markOccurrence(marker: KIVToken): Boolean = marker match {
    case OperationToken(otherop) =>
      op.op == otherop.op
    case OperationDefinitionToken(opdef) =>
      op.op.opsym == opdef.opdefsym && op.op.typ == opdef.typ
    case _ =>
      false
  }
  override val canRename: Boolean = true
}

case class OperationDefinitionToken(val opdef: Opdef) extends KIVToken {
  override val location = opdef.opdefsymloc.loc
  override val eclipseToken: IToken = TokenMap.OP
  override def stringPresentation: String = {
    val domainStr = if(opdef.optdomain.isDefined) {
      "   with   " + xpp(opdef.optdomain.get)
    } else ""
    opdef.opdefsymloc.sym.name + " : " + xpp(opdef.typ) + domainStr
  }
  override val markOccurrences: Boolean = true
  override def markOccurrence(marker: KIVToken): Boolean = marker match {
    case OperationToken(op) =>
      op.opsym == opdef.opdefsym && op.typ == opdef.typ
    case OperationDefinitionToken(otheropdef) if opdef == otheropdef =>
      true
    case _ =>
      false
  }
  override val canRename: Boolean = true
}

case class VariableToken(val prexov: PreAnyXov) extends KIVToken {
  override val location: Location = prexov.location.get
  override val eclipseToken: IToken = TokenMap.VARIABLE
  override def stringPresentation: String = {
    prexov match {
      case prexov: PreXov   => prexov.xov.xovstring
      case poly: PrePolyXov => poly.polyxovsym.name
    }
  }
  override val markOccurrences: Boolean = true
  override def markOccurrence(marker: KIVToken): Boolean = marker match {
    case VariableToken(otherprexov: PreXov) =>
      prexov match {
        case prexov: PreXov =>
          prexov.xov == otherprexov.xov
        case _ =>
          false
      }
    case _ =>
      false
  }
  override val canRename: Boolean = true
}

case class ProcedureToken(val preproc: PreProc) extends KIVToken {
  override val location: Location = preproc.location.get
  override val eclipseToken: IToken = TokenMap.PROC
  override def stringPresentation: String = {
    preproc.proc.procsym.name
  }
  override val markOccurrences: Boolean = true
  override def markOccurrence(marker: KIVToken): Boolean = marker match {
    case ProcedureToken(otherpreproc) =>
      preproc.proc == otherpreproc.proc
    case _ =>
      false
  }
  override val canRename: Boolean = true
}

case class TheoremNameToken(val strloc: StringAndLocation) extends KIVToken {
  override val location: Location = strloc.loc
  override val eclipseToken: IToken = TokenMap.THEOREM
  override def stringPresentation: String = strloc.str
}

case class SpecificationNameToken(val strloc: StringAndLocation) extends KIVToken {
  override val location: Location = strloc.loc
  override val eclipseToken: IToken = TokenMap.SPECIFICATION
  override def stringPresentation: String = strloc.str
}
