package de.isse.kiv.ui.wizards

import java.util.ArrayList
import java.util.zip.ZipEntry

import scala.collection.JavaConverters.asScalaBufferConverter

import org.eclipse.core.resources._
import org.eclipse.core.runtime.CoreException
import org.eclipse.core.runtime.IPath
import org.eclipse.core.runtime.Path
import org.eclipse.jface.dialogs.MessageDialog
import org.eclipse.jface.viewers.IStructuredSelection
import org.eclipse.swt.SWT
import org.eclipse.swt.events.ModifyEvent
import org.eclipse.swt.events.ModifyListener
import org.eclipse.swt.layout.GridData
import org.eclipse.swt.layout.GridLayout
import org.eclipse.swt.widgets._
import org.eclipse.ui.IWorkbench
import org.eclipse.ui.dialogs.FileSystemElement
import org.eclipse.ui.internal.ide.IDEWorkbenchPlugin
import org.eclipse.ui.internal.wizards.datatransfer._

import ArchiveWizardPage._
import de.isse.kiv.project.Builder
import de.isse.kiv.project.Nature
import de.isse.kiv.ui.Console

object ArchiveWizardPage {

  private val EMPTY_FOLDER_MESSAGE = "There are no resources currently selected for import."

  private val DEVGRAPH_MISSING_MESSAGE = "Devgraph missing, or not selected"

  private val STANDARD_PROJECT_NAME = "MyKIVProject"

  private val WIZARD_TITLE = "Import a KIV project archive"

  private val DESTINATION_FIELD_TITLE = "Specify projectname (new project will be created if project does not exist):"

  private val SIZING_TEXT_FIELD_WIDTH = 250;
}

class ArchiveWizardPage(workbench: IWorkbench, selection: IStructuredSelection, fileImportMask: Array[String]) extends WizardArchiveFileResourceImportPage1(workbench, selection, fileImportMask) {

  setTitle(WIZARD_TITLE)

  override def finish(): Boolean = {
    if (!ensureSourceIsValid()) {
      return false
    }
    saveWidgetValues()
    val resourcesEnum = getSelectedResources.iterator()
    val fileSystemObjects = new ArrayList[Object]()
    val devGraphFileOccurrence = new ArrayList[String]()
    while (resourcesEnum.hasNext) {

      val obj = resourcesEnum.next().asInstanceOf[FileSystemElement].getFileSystemObject

      if (obj.isInstanceOf[ZipEntry] && obj.asInstanceOf[ZipEntry].getName.endsWith("devgraph"))
        devGraphFileOccurrence.add(obj.asInstanceOf[ZipEntry].getName)

      else if (obj.isInstanceOf[TarEntry] && obj.asInstanceOf[TarEntry].getName.endsWith("devgraph"))
        devGraphFileOccurrence.add(obj.asInstanceOf[TarEntry].getName)

      fileSystemObjects.add(obj)

    }

    //A valid project archive has only one devgraph. The parentdirectory of the devgraph is seen as the project root 
    //an used to move the files from the archive to the correct folderlevel in the project   

    if (fileSystemObjects.size == 0) {
      MessageDialog.openInformation(getContainer.getShell, DataTransferMessages.DataTransfer_information, DataTransferMessages.FileImport_noneSelected)
      return false
    } else if (devGraphFileOccurrence.size() != 1) {
      MessageDialog.openInformation(getContainer.getShell, DataTransferMessages.DataTransfer_information, "The project archive seems to be broken. The project contains " + devGraphFileOccurrence.size() + " devraph(s)")
      return false
    }

    val importok = importResources(fileSystemObjects)
    if (importok) {

      //Extract the path of the devgraph
      val p: Path = new Path(devGraphFileOccurrence.get(0))

      //If necessary move the project folder (contains devgraph) to root level
      if (!p.removeLastSegments(1).isEmpty)
        repairFolderStructure(p.removeLastSegments(1))

      setProjectNature()
    }
    return importok

  }

  def repairFolderStructure(projectpath: IPath) {
    val containerPath = getContainerFullPath
    val projectPath: IPath = containerPath.removeLastSegments(containerPath.segmentCount() - 1)
    val workspace = IDEWorkbenchPlugin.getPluginWorkspace
    val ip: IProject = workspace.getRoot.getProject(projectPath.toString)

    val path = new Path(sourceNameField.getText)
    val name = path.removeFileExtension.lastSegment

    //Create a unique tmp folder
    var tmpfoldername = "tmpfolder" + System.currentTimeMillis
    var subfolder: IFolder = ip.getFolder(projectpath)
    var tmpfolder: IFolder = ip.getFolder(tmpfoldername)
    while (tmpfolder.exists()) {
      tmpfoldername = "tmpfolder" + System.currentTimeMillis
      tmpfolder = ip.getFolder(tmpfoldername)
    }
    tmpfolder.create(IResource.NONE, true, null);

    if (subfolder.exists()) {

      try {
        //To prevent conflicts, all relevant files are moved into an "unique" folder first
        subfolder.members().foreach {

          case x: IFolder =>
            x.move(projectPath.append(tmpfoldername).append(x.getName), true, null)

          case x: IFile =>
            x.move(projectPath.append(tmpfoldername).append(x.getName), true, null)

          case _ =>
        }

        /*Clean up:
        Delete everthing, except the tmpfolder and the .project file*/
        ip.members().foreach {
          case x: IFolder =>
            if (x.getName != tmpfoldername)
              x.delete(true, false, null)
          case x: IFile =>
            if (x.getName != ".project")
              x.delete(true, false, null)
          case _ =>
        }

        //Move everything from the tmpfolder to the root
        tmpfolder.members().foreach {

          case x: IFolder =>
            x.move(projectPath.append(x.getName), true, false, null)

          case x: IFile =>
            if (x.getName != ".project")
              x.move(projectPath.append(x.getName), true, null)

          case _ =>
        }

        //delete the tmpfolder
        tmpfolder.delete(true, false, null)

        // Fix missing proofs directories (they are missing if they were empty, which happens when the specification is not installed)
        val specsdir = ip.getFolder("specs")
        specsdir.members().foreach {
          case specfolder: IFolder if specfolder.members.forall { _.getName != "proofs" } =>
            specfolder.getFolder("proofs").create(true, true, null)
          case _ =>
        }

      } catch {
        case t: Throwable => Console.error.println(t.toString())
      }
    } else
      Console.error.println("no subfolder")

  }

  def setProjectNature() {
    val containerPath = getContainerFullPath
    val projectPath = containerPath.removeLastSegments(containerPath.segmentCount() - 1)
    val workspace = IDEWorkbenchPlugin.getPluginWorkspace
    val ip = workspace.getRoot.getProject(projectPath.toString)

    try {
      initializeNature(ip)
      ip.setDefaultCharset("UTF-8", null)
    } catch {
      case e: CoreException => e.printStackTrace()
    }
  }

  override def createControl(parent: Composite) {
    initializeDialogUnits(parent)
    val composite = new Composite(parent, SWT.NULL)
    composite.setLayout(new GridLayout())
    composite.setLayoutData(new GridData(GridData.VERTICAL_ALIGN_FILL | GridData.HORIZONTAL_ALIGN_FILL))
    composite.setSize(composite.computeSize(SWT.DEFAULT, SWT.DEFAULT))
    composite.setFont(parent.getFont)
    createSourceGroup(composite)
    createMyDestinationGroup(composite)
    containerNameField.setText(STANDARD_PROJECT_NAME)
    createOptionsGroup(composite)
    restoreWidgetValues()
    updateWidgetEnablements()
    // setPageComplete(determinePageCompletion())
    //setPageComplete(true)
    setErrorMessage(null)
    setControl(composite)

    sourceNameField.addModifyListener(new ModifyListener() {

      def modifyText(e: ModifyEvent) {
        setSuggestedProjectName(sourceNameField.getText)
      }
    });

  }

  def setSuggestedProjectName(zipPath: String) {

    val path = new Path(zipPath)
    val name = path.removeFileExtension.lastSegment.replace(".", "")
    if (name != null)
      containerNameField.setText(name)
  }

  override protected def getResourcePath(): IPath = {
    getPathFromText(this.containerNameField)
  }

  protected def createMyDestinationGroup(parent: Composite) {
    val containerGroup = new Composite(parent, SWT.NONE)
    val layout = new GridLayout()

    layout.numColumns = 1
    containerGroup.setLayout(layout)
    containerGroup.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL))
    containerGroup.setFont(parent.getFont)
    val resourcesLabel = new Label(containerGroup, SWT.NONE)
    resourcesLabel.setText(DESTINATION_FIELD_TITLE)
    resourcesLabel.setFont(parent.getFont)
    containerNameField = new Text(containerGroup, SWT.SINGLE | SWT.BORDER)
    containerNameField.addListener(SWT.Modify, this)
    val data = new GridData(GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL)
    data.widthHint = SIZING_TEXT_FIELD_WIDTH
    containerNameField.setLayoutData(data)
    containerNameField.setFont(parent.getFont)
    initialPopulateContainerField()
  }

  private var containerNameField: Text = _

  private def initializeNature(project: IProject) {
    val description = project.getDescription
    if (!project.hasNature(Nature.NATURE_ID)) {
      val ids = description.getNatureIds
      val newIds = Array.ofDim[String](ids.length + 1)
      System.arraycopy(ids, 0, newIds, 0, ids.length)
      newIds(ids.length) = Nature.NATURE_ID
      description.setNatureIds(newIds)
    }
    var hasBuilder = false
    val cmds = description.getBuildSpec
    for (j <- 0 until cmds.length if cmds(j).getBuilderName == Builder.BUILDER_ID) hasBuilder = true
    if (!hasBuilder) {
      val cmd = description.newCommand()
      cmd.setBuilderName(Builder.BUILDER_ID)
      val newCmds = Array.ofDim[ICommand](cmds.length + 1)
      System.arraycopy(cmds, 0, newCmds, 0, cmds.length)
      newCmds(cmds.length) = cmd
      description.setBuildSpec(newCmds)
    }
    project.setDescription(description, null)
  }

  protected def validateDestination(): Boolean = {
    val containerPath = getContainerFullPath
    if (containerPath == null) {
      setMessage(EMPTY_FOLDER_MESSAGE)
      return false
    }
    true
  }

  private def containsDevgraph(): Boolean = {
    val tmp = getSelectedResources.asScala
    for (`object` <- tmp) {
      setErrorMessage(null)
      // TODO: this is just stupid...
      if (`object`.toString == "File(devgraph)") return true
    }
    setErrorMessage(DEVGRAPH_MISSING_MESSAGE)
    false
  }

  def setPath(path: String) {
    setSourceName(path);
    //sourceNameField.setText(path)
    selectionGroup.setFocus();
    // getContainer().updateButtons()

  }

  def setProjectName(name: String) {
    containerNameField.setText(name)
  }

  override protected def determinePageCompletion(): Boolean = {
    val complete = validateSourceGroup() && containsDevgraph() && validateDestination() && validateOptionsGroup()
    if (complete) {
      setErrorMessage(null)
    }
    complete

  }

  override protected def isPageComplete(): Boolean = {
    determinePageCompletion
  }

}